import pandas as pd
from numpy.core.numeric import nan
from goerr import Err
from .db import Db
from .charts import Plot
from .maps import Map
from .data import Df
from .report import Report
from .messages import Messages


version = "0.4.18"


class DataSwim(Db, Df, Plot, Map, Report, Messages, Err):
    """
    Main class
    """

    def __init__(self, df=None, db=None):
        """
        Initialize with an empty dataframe
        """
        self.msg = Messages()
        self.version = version
        self.df = df
        self.db = db
        self.x = None
        self.y = None
        self.chart_obj = None
        self.chart_opts = dict(width=940)
        self.chart_style = {}
        self.engine = "bokeh"
        self.label = None
        self.backup_df = None
        self.autoprint = True
        self.errors_handling = "exceptions"
        self.notebook = False
        self.header = self._header
        self.footer = self._footer
        self.reports = []
        self.report_engines = [self.engine]
        self.start_time = None
        self.influx_cli = None
        self.datapath = None
        self.report_path = None
        self.static_path = None
        self.quiet = False
        self.nan = nan
        self.color_index = 0
        self.dsmap = None

    def __repr__(self):
        num = 0
        if self.df is not None:
            num = len(self.df.index)
        msg = "<DataSwim object | " + str(num) + " rows>"
        return msg

    def new_(self, df=pd.DataFrame(), db=None, quiet=False):
        """
        Returns a new DataSwim instance from a dataframe
        """
        ds2 = DataSwim(df, db)
        if quiet is False:
            self.ok("A new instance was created")
        return ds2


ds = DataSwim()



# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
from numpy import nan


class Transform():
    """
    Class to transform data
    """

    def keep_(self, *fields):
        """
        Limit a dataframe to some columns
        """
        try:
            ds2 = self._duplicate_(self.df[list(fields)])
        except Exception as e:
            self.err(e, "Can not remove colums")
            return
        self.ok("Columns", " ,".join(fields), "kept")
        return ds2

    def keep(self, *fields):
        """
        Limit a dataframe to some columns
        """
        try:
            self.df = self.df[list(fields)]
        except Exception as e:
            self.err(e, "Can not remove colums")
            return
        self.ok("Setting dataframe to columns", " ".join(fields))

    def drop(self, *cols):
        """
        Drops columns from the main dataframe
        """
        try:
            index = self.df.columns.values
            for col in cols:
                if col not in index:
                    self.warning("Column", col, "not found. Aborting")
                    return
                for col in cols:
                    self.df = self.df.drop(col, axis=1)
        except Exception as e:
            self.err(e, self.drop, "Can not drop column")

    def exclude(self, col, val):
        """
        Delete rows based on value
        """
        try:
            self.df = self.df[self.df[col] != val]
        except Exception as e:
            self.err(e, "Can not exclude rows based on value " + str(val))

    def rsum(self, time_period="1Min", num_col="Number",
             dateindex=None, index_col="Date", fill_col=None):
        """
        Resample, and sum the main dataframe to a time period
        """
        df = self._resample_("sum", time_period,
                                     num_col, dateindex, index_col, fill_col)
        if df is None:
            self.err("Can not sum data")
            return
        self.df = df

    def rsum_(self, time_period="1Min", num_col="Number",
              dateindex=None, index_col="Date", fill_col=None):
        """
        Resample, and sum a dataframe to a time period
        """
        df = self._resample_("sum", time_period,
                            num_col, dateindex, index_col, fill_col)
        if df is None:
            self.err("Can not sum data")
            return
        return self._duplicate_(df)

    def _resample_(self, method, time_period, num_col,
                  dateindex, index_col, fill_col):
        try:
            ds2 = self._duplicate_()
            if dateindex is not None:
                ds2 = self.dateindex_(dateindex)
                if ds2 is None:
                    self.err(e, "Can not process date index")
                    return
            if num_col is not None:
                ds2.add(num_col, 1)
            ds2.df = ds2.df.resample(time_period)
            if method == "sum":
                ds2.df = ds2.df.sum()
            elif method == "mean":
                if num_col is not None:
                    num_vals = ds2.df[num_col].sum()
                ds2.df = ds2.df.mean()
                ds2.df[num_col] = num_vals
            else:
                self.err("Resampling method " + method + " unknown")
                return
            self.ok("Data resampled by", time_period)
            ds2 = self.transform_(None, index_col,
                                 fill_col, None, df=ds2.df)
            df = ds2.df
            if df is None:
                self.err("Can not transform data after resampling")
                return
            return df
        except Exception as e:
            self.err(e, "Can not resample data")
            return

    def rmean(self, time_period="1Min", num_col="Number",
              dateindex=None, index_col="Date", fill_col=None):
        """
        Resample, and sum the main dataframe to a time period
        """
        df = self._resample_("mean", time_period,
                            num_col, dateindex, index_col, fill_col)
        if df is None:
            self.err("Can not mean data")
            return
        self.df = df

    def rmean_(self, time_period="1Min", num_col="Number",
               dateindex=None, index_col="Date", fill_col=None):
        """
        Resample, and sum a dataframe to a time period
        """
        df = self._resample_("mean", time_period,
                            num_col, dateindex, index_col, fill_col)
        if df is None:
            self.err("Can not mean data")
            return
        return self._duplicate_(df)

    def sum_(self, column):
        """
        Returns the sum of all values in a column
        """
        try:
            df = self.df[column]
            num = df.sum()
            return num
        except Exception as e:
            self.err(e, "Can not sum data on column " + str(column))

    def reverse(self):
        """
        Reverses the main dataframe order
        """
        try:
            self.df = self.df.iloc[::-1]
        except Exception as e:
            self.err(e, "Can not reverse the dataframe")

    def sort(self, column):
        """
        Sorts the main dataframe according to the given column
        """
        try:
            self.df = self.df.copy().sort_values(column)
        except Exception as e:
            self.err(e, "Can not sort the dataframe from column " + 
                     str(column))

    def apply(self, function, *cols, **kwargs):
        """
        Apply a function on columns values
        """
        try:
            if len(cols) == 0:
                self.df = self.df.apply(function, **kwargs)
            else:
                cols = list(cols)
                self.df[cols] = self.df[cols].apply(function, **kwargs)
        except Exception as e:
            self.err(e, "Can not apply function")

    def replace(self, col, origin, dest, **kwargs):
        """
        Replace a value in a column
        """
        try:
            self.df[col] = self.df[col].replace(origin, dest)
        except Exception as e:
            self.err(e, "Can not replace value")
            return
        self.ok("Replaced value in column " + str(col))

    def pivot(self, index, **kwargs):
        """
        Pivots a dataframe
        """
        df = self._pivot(index, **kwargs)
        if df is None:
            self.err("Can not pivot table")
            return
        self.df = df

    def pivot_(self, index, **kwargs):
        """
        Pivots a dataframe
        """
        df = self._pivot(index, **kwargs)
        if df is None:
            self.err("Can not pivot table")
            return
        return self._duplicate_(df)

    def _pivot(self, index, **kwargs):
        try:
            kwargs["index"] = index
            return pd.pivot_table(self.df, **kwargs)
        except Exception as e:
            self.err(e, "Can not pivot table")

    def concat(self, *dss, **kwargs):
        """
        Concatenate instances from a list and set it to the main dataframe
        """
        df = self._concat(*dss, **kwargs)
        if df is None:
            self.err("Can not concatenate data")
            return
        self.df = df

    def concat_(self, *dss, **kwargs):
        """
        Concatenate dataframes from a list and returns a new DataSwim instance
        """
        df = self._concat(*dss, **kwargs)
        if df is None:
            self.err("Can not concatenate data")
            return
        return self._duplicate_(df)

    def _concat(self, *dss, **kwargs):
        try:
            df = pd.DataFrame()
            for dsx in dss:
                df = pd.concat([df, dsx.df], **kwargs)
            return df
        except Exception as e:
            self.err(e, "Can not concatenate data")

    def split_(self, col):
        """
        Split the main dataframe according to a column's unique values and
        return a dict of dataswim instances
        """
        try:
            dss = {}
            unique = self.df[col].unique()
            for key in unique:
                df2 = self.df.loc[self.df[col] == key]
                ds2 = self._duplicate_(df2)
                dss[key] = ds2
            return dss
        except Exception as e:
            self.err(e, "Can not split dataframe")

    def add(self, column, value):
        """
        Add a columns with default values
        """
        try:
            df = self.df.copy()
            df[column] = value
            self.df = df
        except Exception as e:
            self.err(e, self.add, "Can not add column")

    def copy_col(self, origin_col, end_col):
        """
        Copy a columns values in another column
        """
        try:
            df = self.df.copy()
            df[end_col] = df[[origin_col]]
            self.df = df
        except Exception as e:
            self.err(e, self.copy_col, "Can not copy column")

    def index_col(self, column="index"):
        """
        Add a column filled from the index to the main dataframe
        """
        try:
            self.df = self._index_col(column)
        except Exception as e:
            self.err(e, self.index_col, "Can not add index column")

    def index_col_(self, column="index"):
        """
        Returns a DatasWim instance with a new column filled from the index
        """
        try:
            df = self._index_col(column)
            return self.df
        except Exception as e:
            self.err(e, self.index_col_, "Can not add index column")

    def _index_col(self, column):
        """
        Add a column from the index
        """
        df = self.df.copy()
        try:
            df[column] = df.index.values
            if self.autoprint is True:
                self.ok("Column", column, "added from index")
            return df
        except Exception as e:
            self.err(e)

    def drop_rows(self, *rows):
        """
        Drops some rows from the main dataframe
        """
        try:
            self.df = self._drop_rows(rows)
        except Exception as e:
            self.err(e, self.drop_rows, "Can not drop rows")
            return

    def _drop_rows(self, rows):
        """
        Drops some rows from the main dataframe
        """
        try:
            df = self.df.copy()
            rows = list(rows)
            df = df.drop(rows)
            return df
        except Exception as e:
            self.err(e, self._drop_rows, "Can not drop rows")
            return
        if self.autoprint is True:
            self.ok("Dropped rows", ", ".join(rows))

    def append(self, vals, index=None):
        """
        Append a row to the main dataframe
        """
        try:
            self.set(self._append(vals, index))
        except Exception as e:
            self.err(e, self.append, "Can not append row")
            return

    def append_(self, vals, index=None):
        """
        Returns a dataswim instance with one more row
        """
        try:
            return self.duplicate_(self._append(vals, index))
        except Exception as e:
            self.err(e, self.append_, "Can not append row")
            return

    def _append(self, vals, index):
        """
        Returns a dataframe with a new row
        """
        try:
            df = self.df.copy()
            if index is None:
                index = len(df)
            df.loc[index] = vals
            return df
        except Exception as e:
            self.err(e, self._append, "Can not append row")
            return
        if self.autoprint is True:
            self.ok("Row added to dataframe")

    def merge(self, df, on, how="outer", **kwargs):
        """
        Set the main dataframe from the current dataframe and the passed
        dataframe
        """
        try:
            df = self._merge(df, on, how, **kwargs)
            self.set(df)
        except Exception as e:
            self.err(e, self.merge, "Can not merge dataframes")

    def merge_(self, df, on, how="outer", **kwargs):
        """
        Returns a Dataswim instance from the current dataframe and the passed
        dataframe
        """
        try:
            df = self._merge(df, on, how, **kwargs)
            ds = self.clone_(df)
            return ds
        except Exception as e:
            self.err(e, self.merge_, "Can not merge dataframes")

    def _merge(self, df, on, how, **kwargs):
        """
        Returns a dataframe from the current dataframe and the passed
        dataframe
        """
        try:
            df = pd.merge(self.df, df, on=on, how=how, **kwargs)
            return df
        except Exception as e:
            self.err(e, self._merge, "Can not merge dataframes")

    def diffn(self, diffcol, name="Diff"):
        """
        Add a diff column to the main dataframe
        """
        try:
            vals = []
            i = 0
            for _, row in self.df.iterrows():
                current = row[diffcol]
                try:
                    nextv = self.df[diffcol].iloc[i + 1]
                except Exception:
                    vals.append(nan)
                    continue
                val = nextv - current
                vals.append(round(val, 1))
                i += 1
            self.add("Diff", vals)
        except Exception as e:
            self.err(e, self._append, "Can not diff column")
            return
        if self.autoprint is True:
            self.ok("Diff column " + name + " added to the dataframe")

    def diff(self, diffcol, name="Diff"):
        """
        Add a diff column to the main dataframe
        """
        try:
            df = self.df.copy()
            previous = 0
            i = 0
            vals = [df[diffcol].iloc[0]]
            for _, row in df.iterrows():
                val = row[diffcol] - previous
                new = round(val, 1)
                previous = row[diffcol]
                if i == 0:
                    vals = [0]
                else:
                    vals.append(new)
                i = 1
            self.df = df
            self.add(name, vals)
        except Exception as e:
            self.err(e, self._append, "Can not diff column")
            return
        if self.autoprint is True:
            self.ok("Diff column " + name + " added to the dataframe")

    def diffm(self, diffcol, name="Diff", default=nan):
        """
        Add a diff from mean column to the main dataframe
        """
        try:
            df = self.df.copy()
            mean = self.df[diffcol].mean()
            vals = []
            for _, row in self.df.iterrows():
                num = row[diffcol]
                if num > 0:
                    diff = int(((num - mean) * 100) / mean)
                    vals.append(diff)
                else:
                    vals.append(default)
            self.df = df
            self.add(name, vals)
        except Exception as e:
            self.err(e, self._append, "Can not diff column")
            return
        if self.autoprint is True:
            self.ok("Diff column " + name + " added to the dataframe")

    def gmean_(self, col, index_col=True):
        """
        Group by and mean column
        """
        try:
            df = self.df.copy()
            df = df.groupby([col]).mean()
            if index_col is True:
                df[col] = df.index.values
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.gmean_, "Can not meansum column")

    def gsum_(self, col, index_col=True):
        """
        Group by and sum column
        """
        try:
            df = self.df.copy()
            df = df.groupby([col]).sum()
            if index_col is True:
                df[col] = df.index.values
            return self.clone_(df)
        except Exception as e:
            self.err(e, self.gsum_, "Can not groupsum column")

    def ratio(self, col, ratio_col="Ratio"):
        """
        Add a column whith the percentages ratio from a column
        """
        try:
            df = self.df.copy()
            df[ratio_col] = df[[col]].apply(
                lambda x: 100 * x / float(x.sum()))
            self.set(df)
        except Exception as e:
            self.err(e, self.ratio, "Can not calculate ratio")

    def rename(self, source_col, dest_col):
        """
        Renames a column in the main dataframe
        """
        try:
            self.df = self.df.rename(columns={source_col: dest_col})
        except Exception as e:
            self.err(e, self.rename, "Can not rename column")
            return
        if self.autoprint is True:
            self.ok("Column", source_col, "renamed")

    def trimquants_(self, col, inf, sup):
        """
        Remove superior and inferior quantiles from the dataframe and
        returns a Dataswim instance
        """
        try:
            ds2 = self.clone_()
            ds2.df = self._trimquants(col, inf, sup)
            return ds2
        except Exception as e:
            self.err(e, self.trimquants_, "Can not trim quantiles")

    def trimquants(self, col, inf, sup):
        """
        Remove superior and inferior quantiles from the dataframe
        """
        try:
            self.set(self._trimquants(col, inf, sup))
        except Exception as e:
            self.err(e, self.trimquants, "Can not trim quantiles")

    def trimsquants(self, col, sup):
        """
        Remove superior quantiles from the dataframe
        """
        try:
            self.set(self._trimquants(col, None, sup))
        except Exception as e:
            self.err(e, self.trimsquants, "Can not trim superior quantiles")

    def trimiquants(self, col, inf):
        """
        Remove superior and inferior quantiles from the dataframe
        """
        try:
            self.set(self._trimquants(col, inf, None))
        except Exception as e:
            self.err(e, self.trimiquants, "Can not trim inferior quantiles")

    def _trimquants(self, col, inf, sup):
        """
        Remove superior and inferior quantiles from the dataframe
        and returs a dataframe
        """
        try:
            ds2 = self.clone_()
            if inf is not None:
                qi = ds2.df[col].quantile(inf)
                ds2.df = ds2.df[ds2.df[col] > qi]
            if sup is not None:
                qs = ds2.df[col].quantile(sup)
                ds2.df = ds2.df[ds2.df[col] < qs]

        except Exception as e:
            self.err(e, self._trimquants, "Can not trim quantiles")
        if self.autoprint is True:
            msg = "Removed values "
            if inf is not None:
                msg += "under " + str(qi)
            if sup is not None and inf is not None:
                msg += " and"
            if sup is not None:
                msg += "upper " + str(qs)
            self.ok(msg, "in column", col)
        return ds2.df

    def transform_(self, dateindex=None, index_col=None,
                   fill_col=None, num_col=None, df=None):
        """
        Returns a DataSwim instance transformed according to the
        given parameters
        """
        if df is None:
            if self.df is None:
                self.err("No dataframe: please provide one "
                         "in parameters or set it")
                return
            df = self.df
        ds2 = self._duplicate_(df)
        if dateindex is None and index_col is None and fill_col \
                is None and num_col is None:
            return ds2
        try:
            if dateindex is not None:
                ds2.dateindex(dateindex, df=df)
            if fill_col is not None:
                ds2.fill_nan(0, fill_col)
            if index_col is not None:
                ds2.index_col(index_col)
            if num_col is not None:
                ds2.add(num_col, 1)
            return ds2
        except Exception as e:
            self.err(e, "Can not transform data")



# -*- coding: utf-8 -*-
import datetime
import pandas as pd
import numpy as np
from numpy import nan
from pandas.testing import assert_frame_equal
from dataswim.tests.base import BaseDsTest
from dataswim import ds

ds.errs_traceback = False


class TestDsDataClean(BaseDsTest):

    def test_drop_nan(self):
        df1 = pd.DataFrame({"one": ["one","two"], "two":
            ["two", None]}, ["1", "2"])
        ds.df = df1
        ds.drop_nan("two")
        df2 = pd.DataFrame({"one": "one", "two": "two"}, ["1"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds.drop_nan(method="any")
        assert_frame_equal(ds.df, df2)
        ds.df = None
        self.assertErr("TypeError", ds.drop_nan, "one")

    def test_zero_nan(self):
        df1 = pd.DataFrame({"one": ["one","two"], "two":
            ["two", 0]}, ["1", "2"])
        ds.df = df1
        ds.zero_nan("two")
        df2 = pd.DataFrame({"one": ["one", "two"], "two":
                ["two", nan]}, ["1", "2"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.zero_nan_("two")
        assert_frame_equal(ds2.df, df2)
        ds.zero_nan("two", "one")
        ds.df = None
        self.assertErr("AttributeError", ds.zero_nan, "one", "two")
        ds.df = df1
        msg = "Column wrong does not exist"
        self.assertWarning(msg, ds.zero_nan, "wrong", "two")
        self.assertErr(None, ds.zero_nan, "wrong", "wrong")
        self.assertWarning(msg, ds.zero_nan_, "wrong", "two")
        self.assertErr(None, ds.zero_nan_, "wrong", "wrong")

    def test_fill_nan(self):
        df1 = pd.DataFrame({"one": ["one", "two"], "two":
                            ["two", None]}, ["1", "2"])
        ds.df = df1
        ds.fill_nan("two", "two")
        df2 = pd.DataFrame({"one": ["one", "two"], "two":
                            ["two", "two"]}, ["1", "2"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds.fill_nan("two")
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.fill_nan_("two", "two")
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        self.assertErr("AttributeError", ds.fill_nan, "val", "two")
        self.assertErr("AttributeError", ds.fill_nan_, "val", "two")
        ds.df = df1
        msg = "Can not find column wrong"
        self.assertWarning(msg, ds.fill_nan, "val", "wrong")
        self.assertErr(None, ds.fill_nan, "val", "wrong")

    def test_replace(self):
        df1 = pd.DataFrame({"one": ["one", "two"], "two":
                            ["two", "two"]}, ["1", "2"])
        ds.df = df1
        ds.replace("one", "two", "three")
        df2 = pd.DataFrame({"one": ["one", "three"], "two":
                            ["two", "two"]}, ["1", "2"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds2 = ds.replace_("one", "two", "three")
        assert_frame_equal(ds2.df, df2)
        ds.df = None
        self.assertErr("AttributeError", ds.replace, "one", "two", "three")
        self.assertErr("AttributeError", ds.replace_, "one", "two", "three")

    def test_to_int(self):
        df1 = pd.DataFrame({"one": ["one", "two"], "two":
                            [1.0, 2.0]}, ["1", "2"])
        ds.df = df1
        ds.to_int("two")
        self.assertEqual(list(ds.df["two"]), [1, 2])
        ds.df = None
        self.assertErr("TypeError", ds.to_int, "two")
        ds.df = pd.DataFrame({"one": ["one", "two"], "two":
                            ["wrong", 2.0]}, ["1", "2"])
        self.assertErr("ValueError", ds.to_int, "two")

    def test_to_float(self):
        df1 = pd.DataFrame({"one": ["one", "two"], "two":
                            [1, 2]}, ["1", "2"])
        ds.df = df1
        ds.to_float("two")
        self.assertEqual(list(ds.df["two"]), [1.0, 2.0])
        ds.df = None
        self.assertErr("AttributeError", ds.to_float, "two")
        ds.df = pd.DataFrame({"one": ["one", "two"], "two":
                            ["wrong", 2]}, ["1", "2"])
        self.assertErr("ValueError", ds.to_float, "two")

    def test_to_type(self):
        df1 = pd.DataFrame({"one": ["one", "two"], "two":
                            [1, 2]}, ["1", "2"])
        ds.df = df1
        df = ds.to_type("float64", "two")
        df2 = pd.DataFrame({"one": ["one", "two"], "two":
                            [1.0, 2.0]}, ["1", "2"])
        assert_frame_equal(df, df2)
        ds.df = None
        self.assertErr("AttributeError", ds.to_type, "float64", "two")
        ds.df = pd.DataFrame({"one": ["one", "two"], "two":
                            ["wrong", 2.0]}, ["1", "2"])
        self.assertErr(None, ds.to_type, "float64", "two")
        self.assertErr(None, ds.to_type, "float64", "wrong")

    def tests_timestamps(self):
        df1 = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2002/12/02"]}, [1, 2])
        ds.df = df1
        ds.timestamps("two",name= "ts")
        df2 = pd.DataFrame({"one": ["one","two"], "ts": [1038700800,
            1038787200], "two":
            ["2002/12/01", "2002/12/02"]}, [1, 2])
        ds.df = ds.df.reindex(sorted(ds.df.columns), axis=1)
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds.timestamps("two")
        ds.timestamps("two", unit="s", errors="raise")
        ds.df = None
        ds.timestamps("two")
        self.assertRaises(AttributeError)

    def test_date_index(self):
        df1 = pd.DataFrame({"one": ["one","two"], "date":
            ["2002/12/01", "2002/12/02"]}, [1, 2])
        ds.df = df1
        ds.dateindex("date")
        index = np.array(['2002-12-01T00:00:00.000000000',
                       '2002-12-02T00:00:00.000000000'],
                        dtype="datetime64[ns]")
        np.testing.assert_array_equal(ds.df.index.values, index)
        ds.df = df1
        ds2 = ds.dateindex_("date")
        np.testing.assert_array_equal(ds2.df.index.values, index)
        ds.df = None
        ds.dateindex_("date")
        self.assertRaises(TypeError)
        ds.dateindex("date")
        self.assertRaises(TypeError)
        ds.dateindex("wrong")
        self.assertRaises(TypeError)

    def test_index(self):
        df1 = pd.DataFrame([[1, 3], [2, 4]], columns=["one", "two"])
        ds.df = df1
        ds.index("one")
        index = np.array([1, 2])
        np.testing.assert_array_equal(ds.df.index.values, index)
        ds.df = df1
        ds2 = ds.index_("one")
        np.testing.assert_array_equal(ds2.df.index.values, index)
        ds.df = None
        ds.index("one")
        self.assertRaises(TypeError)
        ds.index_("one")
        self.assertRaises(TypeError)

    def test_strip(self):
        df1 = pd.DataFrame([[" whitespace ", 3], [2, 4]],
                           index=["one", "two"])
        ds.df = df1
        ds.strip("one")
        df2 = pd.DataFrame([["whitespace", 3], [2, 4]],
                           index=["one", "two"])
        assert_frame_equal(ds.df, df2)
        ds.df = None
        ds.strip("one")
        self.assertRaises(TypeError)
        ds.df = "wrong"
        ds.index("one")
        self.assertRaises(AttributeError)

    def test_strip_cols(self):
        df1 = pd.DataFrame([[1, 3], [2, 4]],
                           columns=[" one ", " two "])
        ds.df = df1
        ds.strip_cols()
        df2 = pd.DataFrame([[1, 3], [2, 4]],
                           columns=["one", "two"])
        assert_frame_equal(ds.df, df2)
        ds.df = pd.DataFrame([[1, 3], [2, 4]],
                           columns=["one", None])
        ds.strip_cols()

    def test_roundvals(self):
        df1 = pd.DataFrame([[1.345854, 3.0], [2.1, 4.0]],
                           columns=["one", "two"])
        ds.df = df1
        ds.roundvals("one")
        df2 = pd.DataFrame([[1.35, 3.0], [2.1, 4.0]],
                           columns=["one", "two"])
        assert_frame_equal(ds.df, df2)
        ds.df = df1
        ds.roundvals("wrong")
        self.assertRaises(KeyError)

    def test_format_date(self):
        date = datetime.datetime(2011, 1, 3, 0, 0)
        d2 = ds.format_date_(date)
        self.assertEqual(d2, '2011-01-03 00:00:00')

    def test_dates(self):
        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2003/12/02"]}, [1, 2])
        ds.date("two")
        df2 = pd.DataFrame({"one": ["one","two"],"two":
            ["2002-12-01 00:00:00", "2003-12-02 00:00:00"]}, [1, 2])
        assert_frame_equal(ds.df, df2)

        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2003/12/02"]}, [1, 2])
        df2 = pd.DataFrame({"one": ["one","two"],"two":
            ["2002", "2003"]}, [1, 2])
        ds.date("two", precision="Y")
        assert_frame_equal(ds.df, df2)

        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2003/12/02"]}, [1, 2])
        df2 = pd.DataFrame({"one": ["one","two"],"two":
            ["2002-12", "2003-12"]}, [1, 2])
        ds.date("two", precision="M")
        assert_frame_equal(ds.df, df2)

        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2003/12/02"]}, [1, 2])
        df2 = pd.DataFrame({"one": ["one","two"],"two":
            ["2002-12-01", "2003-12-02"]}, [1, 2])
        ds.date("two", precision="D")
        assert_frame_equal(ds.df, df2)

        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2003/12/02"]}, [1, 2])
        df2 = pd.DataFrame({"one": ["one","two"],"two":
            ["2002-12-01 00", "2003-12-02 00"]}, [1, 2])
        ds.date("two", precision="H")
        assert_frame_equal(ds.df, df2)

        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2003/12/02"]}, [1, 2])
        df2 = pd.DataFrame({"one": ["one","two"],"two":
            ["2002-12-01 00:00", "2003-12-02 00:00"]}, [1, 2])
        ds.date("two", precision="Min")
        assert_frame_equal(ds.df, df2)

        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "2003/12/02"]}, [1, 2])
        ds.date("wrong")
        self.assertRaises(KeyError)

        ds.df = pd.DataFrame({"one": ["one","two"], "two":
            ["2002/12/01", "wrong"]}, [1, 2])
        ds.date("two", "wrong")
        self.assertRaises(TypeError)

        ds.df = None
        ds.date("two")
        self.assertRaises(ValueError)

    def test_nulls(self):
        df1 = pd.DataFrame({"one": ["one", "two"], "two":
                            ["", None]}, ["1", "2"])
        ds.df = df1
        ds.fill_nulls("two")
        df2 = pd.DataFrame({"one": ["one", "two"], "two":
                            [nan, nan]}, ["1", "2"])
        assert_frame_equal(ds.df, df2)
        ds.df = None
        ds.fill_nulls("two")
        self.assertRaises(AttributeError)


    def test_nan_empty(self):
        df1 = pd.DataFrame({"one": ["one","two"], "two":
            ["two", ""]}, ["1", "2"])
        ds.df = df1
        ds.nan_empty("two")
        df2 = pd.DataFrame({"one": ["one","two"], "two":
            ["two", nan]}, ["1", "2"])
        assert_frame_equal(ds.df, df2)
        ds.df = None
        ds.nan_empty("two")
        self.assertRaises(TypeError)




# -*- coding: utf-8 -*-
import time
import arrow
import pandas as pd
from numpy.core.numeric import nan
from goerr import Err
from goerr.colors import colors


class Clean(Err):
    """
    Class to clean the data
    """

    def drop_nan(self, field=None, method="all", **kwargs):
        """
        Drop NaN values from the main dataframe
        """
        try:
            if field is None:
                self.df = self.df.dropna(how=method, **kwargs)
            else:
                self.df = self.df[self.df[field].notnull()]
        except Exception as e:
            self.error(e, "Error dropping nan values")

    def nan_empty(self, field):
        """
        Fill empty values with NaN values
        """
        try:
            self.df[field] = self.df[field].replace('', nan)
            self.ok("Filled empty values with nan in column " + field)
        except Exception as e:
            self.err(e, "Can not fill empty values with nan")

    def zero_nan(self, *fields):
        """
        Converts zero values to nan values in selected columns
        """
        df = self._zero_nan(*fields)
        if df is None:
            self.err("Can not fill zero values with nan")
            return
        self.df = df

    def zero_nan_(self, *fields):
        """
        Returns a DataSwim instance with zero values to nan values in
        selected columns
        """
        df = self._zero_nan(*fields)
        if df is None:
            self.err("Can not fill zero values with nan")
            return
        ds2 = self.clone_(quiet=True)
        ds2.df = df
        return ds2

    def _zero_nan(self, *fields):
        try:
            df = self.df.copy()
            for field in fields:
                if field not in df.columns:
                    self.warning("Column "+field+" does not exist")
                    return
                df = df.replace({field: {0: nan}})
            if len(fields) > 1:
                self.ok("Replaced 0 values by nan in columns",
                        str(fields))
            else:
                self.ok("Replaced 0 values by nan in column",
                        field)
            return df
        except Exception as e:
            self.err(e)

    def fill_nan(self, val, *fields):
        """
        Fill NaN values with new values in the main dataframe
        """
        df = self._fill_nan(val, *fields)
        if df is not None:
            self.df = df
        else:
            self.err("Can not fill nan values")

    def fill_nan_(self, val, *fields):
        """
        Returns a DataSwim instance with NaN values filled
        """
        df = self._fill_nan(val, *fields)
        if df is not None:
            return self._duplicate_(df=df)
        else:
            self.err("Can not fill nan values")

    def _fill_nan(self, val, *fields):
        fields = list(fields)
        if len(fields) == 0:
            fields = self.df.columns.values
        df = None
        try:
            df = self.df.copy()
            for el in fields:
                try:
                    df[el] = df[[el]].fillna(val)
                except KeyError:
                    self.warning("Can not find column", el)
                    return
        except Exception as e:
            self.err(e, "Can not fill nan values")
            return
        self.ok("Filled nan values in columns", ",".join(fields))
        return df

    def replace(self, col, searchval, replaceval):
        """
        Replace a value in a column in the main dataframe
        """
        df = self._replace(col, searchval, replaceval)
        if df is not None:
            self.df = df
        else:
            self.err("Can not replace value in column")

    def replace_(self, col, searchval, replaceval):
        """
        Returns a Dataswim instance with replaced values in a column
        """
        df = self._replace(col, searchval, replaceval)
        if df is not None:
            return self._duplicate_(df)
        else:
            self.err("Can not replace value in column")

    def _replace(self, col, searchval, replaceval):
        try:
            df = self.df.copy()
            df[col] = df[col].replace(searchval, replaceval)
            return df
        except Exception as e:
            self.err(e, "Can not replace value in column")

    def fill_nulls(self, field):
        """
        Fill all null values with NaN values
        """
        n = [None, ""]
        try:
            self.df[field] = self.df[field].replace(n, nan)
        except Exception as e:
            self.err(e)

    def to_int(self, field):
        """
        Convert some column values to integers
        """
        def convert(val):
            return int(val)

        try:
            self.df[field] = self.df[field].apply(convert)
        except Exception as e:
            self.err(e, "Can not convert column values to integer")
            return
        self.ok("Converted column values to integers")

    def to_float(self, *cols):
        """
        Convert colums values to float
        """
        df = self.to_type("float64", *cols)
        if df is not None:
            self.df = df
        else:
            self.err("Can not convert column values to float")
            return
        self.ok("Converted column values to float")

    def to_type(self, dtype, *cols):
        """
        Convert colums values to a given type
        """
        try:
            df = self.df.copy()
            allcols = df.columns.values
            for col in cols:
                if col not in allcols:
                    self.err("Column " + col + " not found")
                    return
                df[col] = df[col].astype(dtype)
            return df
        except Exception as e:
            self.err(e, "Can not convert to type")

    def timestamps(self, col, **kwargs):
        """
        Add  a timestamps column from a date column
        """
        try:
            name = "Timestamps"
            if "name" in kwargs:
                name = kwargs["name"]
            if "errors" not in kwargs:
                kwargs["errors"] = "coerce"
            if "unit" in kwargs:
                kwargs["unit"] = "ms"
            try:
                self.df[col] = pd.to_datetime(self.df[col], **kwargs)
            except TypeError:
                pass
            ts = []
            for el in self.df[col]:
                ts.append(arrow.get(el).timestamp)
            self.df[name] = ts
        except Exception as e:
            self.err(e, "Can not convert to timestamps")

    def date(self, *fields, precision="S", format=None):
        """
        Convert column values to properly formated datetime
        """
        def formatdate(row):
            return row.strftime(format)
            
        def convert(row):
            encoded = '%Y-%m-%d %H:%M:%S'
            if precision == "Min":
                encoded = '%Y-%m-%d %H:%M'
            elif precision == "H":
                encoded = '%Y-%m-%d %H'
            elif precision == "D":
                encoded = '%Y-%m-%d'
            elif precision == "M":
                encoded = '%Y-%m'
            elif precision == "Y":
                encoded = '%Y'
            return row.strftime(encoded)

        try:
            for f in fields:
                try:
                    if format is None:
                        self.df[f] = pd.to_datetime(self.df[f]).apply(convert)
                    else:
                        self.df[f] = pd.to_datetime(self.df[f]).apply(formatdate)
                except ValueError as e:
                    self.err(e, "Can not convert date")
                    return
        except KeyError:
            self.warning("Can not find colums " + " ".join(fields))
            return
        except Exception as e:
            self.err(e, "Can not process date field")

    def dateindex(self, datafield):
        """
        Set a datetime index from a column
        """
        df = self._dateindex(datafield)
        if df is None:
            self.err("Can not index data")
            return
        self.df = df

    def dateindex_(self, datafield):
        """
        Returns a DataSwim instance index from a column
        """
        df = self._dateindex(datafield)
        if df is None:
            self.err("Can not index data")
            return
        return self._duplicate_(df)

    def _dateindex(self, datafield):
        try:
            index = pd.DatetimeIndex(self.df[datafield])
            df = self.df.set_index(index)
        except Exception as e:
            self.err(e, "Can not index with column " +
                     colors.bold(datafield))
            return
        self.ok("Added a datetime index from column", datafield)
        return df

    def index(self, indexcol):
        """
        Set an index to the main dataframe
        """
        df = self._index(indexcol)
        if df is None:
            self.err("Can not create index")
            return
        self.df = df

    def index_(self, indexcol):
        """
        Returns a Dataswim instance with an index
        """
        df = self._index(indexcol)
        if df is None:
            self.err("Can not create index")
            return
        return self._duplicate_(df)

    def _index(self, indexcol):
        try:
            df = self.df.set_index(self.df[indexcol])
        except Exception as e:
            self.err(e, self._index, "Can not find column " +
                     colors.bold(indexcol) + " in data")
            return
        self.ok("Added an index from column", indexcol)
        return df

    def strip(self, col):
        """
        Remove leading and trailing white spaces in a column's values
        """
        def remove_ws(row):
            val = str(row[col])
            if " " in val:
                row[col] = val.strip()
            return row

        try:
            self.df.apply(remove_ws)
        except Exception as e:
            self.err(e, "Can not remove white space in column")
            return
        self.ok("White space removed in column values")

    def strip_cols(self):
        """
        Remove leading and trailing white spaces in columns names
        """
        cols = {}
        skipped = []
        for col in self.df.columns.values:
            try:
                cols[col] = col.strip()
            except Exception:
                skipped.append(str(col))
        self.df = self.df.rename(columns=cols)
        self.ok("White spaces removed in columns names")
        if len(skipped) > 0:
            self.info("Skipped columns", ','.join(
                skipped), "while removing white spaces")

    def roundvals(self, col, precision=2):
        """
        Round floats in a column
        """
        try:
            self.df[col] = self.df[col].astype("float64")
            self.df[col] = self.df[col].apply(lambda x: round(x, precision))
        except Exception as e:
            self.err(e, "Can not round column values")
            return
        self.ok("Rounded values in column " + col)

    def format_date_(self, date):
        """
        Formats a date
        """
        return date.strftime('%Y-%m-%d %H:%M:%S')


# -*- coding: utf-8 -*-
import pandas as pd
import deepdish as dd
from .views import View
from .clean import Clean
from .count import Count
from .select import Select
from .transform import Transform
from .export import Export
from .search import Search
from .stats import Stats
from .text import Text


class Df(Select, View, Transform, Clean, Count, Export, Search, Stats, Text):
    """
    Class for manipulating dataframes
    """

    def _duplicate_(self, df=None, db=None, quiet=True):
        try:
            if db is None:
                db = self.db
            if df is None:
                if self.df is None:
                    self.err("The main dataframe is empty and no dataframe"
                             " was provided: please provide a dataframe as argument")
                    return
                df = self.df.copy()
            ds2 = self.new_(df, db, True)
            ds2.db = self.db
            ds2.x = self.x
            ds2.y = self.y
            ds2.chart_obj = self.chart_obj
            ds2.chart_opts = self.chart_opts
            ds2.chart_style = self.chart_style
            ds2.label = self.label
            ds2.reports = self.reports
            ds2.report_engines = self.report_engines
            ds2.backup_df = self.backup_df
            ds2.autoprint = self.autoprint
            ds2.errors_handling = self.errors_handling
            ds2.datapath = self.datapath
            ds2.report_path = self.report_path
            ds2.static_path = self.static_path
        except Exception as e:
            self.err(e, "Can not duplicate instance")
            return
        if quiet is False:
            self.ok("A duplicated instance was created")
        return ds2

    def clone_(self, quiet=False):
        """
        Clone the DataSwim instance
        """
        ds2 = self._duplicate_(quiet=True)
        if ds2 is None:
            self.err("Can not clone instance")
        else:
            if quiet is False:
                self.ok("Instance cloned")
        return ds2

    def backup(self):
        """
        Backup the main dataframe
        """
        try:
            self.backup_df = self.df.copy()
        except Exception as e:
            self.err(e, "Can not backup data")
            return
        self.msg.ok("Dataframe backed up")

    def restore(self):
        """
        Restore the main dataframe
        """
        if self.backup_df is None:
            self.warning("No dataframe is backed up: nothing restore")
            return
        self.df = self.backup_df
        self.msg.ok("Dataframe is restored")

    def load_json(self, path, **kwargs):
        """
        Load data in the main dataframe from json
        """
        try:
            df = pd.read_json(path, **kwargs)
            self.df = df
        except Exception as e:
            self.err(e, "Can not load json")

    def load_h5(self, filepath):
        """
        Load a Hdf5 file to the main dataframe
        """
        try:
            self.start("Loading Hdf5 data...")
            self.df = dd.io.load(filepath)
            self.end("Finished loading Hdf5 data")
        except Exception as e:
            self.err(e, "Can not load Hdf5 file")

    def load_excel(self, filepath, **kwargs):
        """
        Set the main dataframe with the content of an Excel file
        """
        try:
            df = pd.read_excel(filepath, **kwargs)
            if len(df.index) == 0:
                self.warning("Empty Excel file. Can not set the dataframe.")
                return
            self.df = df
        except Exception as e:
            self.err(e, "Can not load Excel file")

    def load_csv(self, url, dateindex=None,
                 index_col=None, fill_col=None, num_col=None, **kwargs):
        """
        Loads csv data in the main dataframe
        """
        self.start("Loading csv...")
        try:
            if self.datapath is not None and url.startswith("/") is False:
                url = self.datapath + "/" + url
            df = pd.read_csv(url, **kwargs)
            self.df = df
        except FileNotFoundError:
            msg = "File " + url + " not found"
            self.warning(msg)
            return
        except Exception as e:
            self.err(e, "Can not load csv file")
            return
        ds2 = self.transform_(dateindex, index_col, fill_col, df=self.df)
        if ds2 is None:
            self.err("Can not load csv")
            return
        self = ds2
        self.end("Finished loading csv")

    def dateparser(self, dformat='%d/%m/%Y'):
        """
        Returns a date parser for pandas
        """
        def dateparse(dates):
            return [pd.datetime.strptime(d, dformat) for d in dates]
        return dateparse
